#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import timedelta
from os.path import isabs

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_wtf.csrf import CSRFProtect
from sqlalchemy.engine.url import make_url

from airflow import settings
from airflow.api_fastapi.app import get_auth_manager
from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException
from airflow.logging_config import configure_logging
from airflow.providers.fab.www.extensions.init_appbuilder import init_appbuilder
from airflow.providers.fab.www.extensions.init_jinja_globals import init_jinja_globals
from airflow.providers.fab.www.extensions.init_manifest_files import configure_manifest_files
from airflow.providers.fab.www.extensions.init_security import init_api_auth
from airflow.providers.fab.www.extensions.init_session import init_airflow_session_interface
from airflow.providers.fab.www.extensions.init_views import (
    init_api_auth_provider,
    init_api_error_handlers,
    init_error_handlers,
    init_plugins,
)
from airflow.providers.fab.www.extensions.init_wsgi_middlewares import init_wsgi_middleware
from airflow.providers.fab.www.utils import get_session_lifetime_config

app: Flask | None = None

# Initializes at the module level, so plugins can access it.
# See: /docs/plugins.rst
csrf = CSRFProtect()


def create_app(enable_plugins: bool):
    """Create a new instance of Airflow WWW app."""
    from airflow.providers.fab.auth_manager.fab_auth_manager import FabAuthManager

    flask_app = Flask(__name__)
    flask_app.secret_key = conf.get("api", "SECRET_KEY")
    flask_app.config["SQLALCHEMY_DATABASE_URI"] = conf.get("database", "SQL_ALCHEMY_CONN")
    flask_app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
    flask_app.config["PERMANENT_SESSION_LIFETIME"] = timedelta(minutes=get_session_lifetime_config())

    flask_app.config["SESSION_COOKIE_HTTPONLY"] = True
    if conf.has_option("fab", "COOKIE_SECURE"):
        flask_app.config["SESSION_COOKIE_SECURE"] = conf.getboolean("fab", "COOKIE_SECURE")
    if conf.has_option("fab", "COOKIE_SAMESITE"):
        flask_app.config["SESSION_COOKIE_SAMESITE"] = conf.get("fab", "COOKIE_SAMESITE")

    webserver_config = conf.get_mandatory_value("fab", "config_file")
    # Enable customizations in webserver_config.py to be applied via Flask.current_app.
    with flask_app.app_context():
        flask_app.config.from_pyfile(webserver_config, silent=True)

    url = make_url(flask_app.config["SQLALCHEMY_DATABASE_URI"])
    if url.drivername == "sqlite" and url.database and not isabs(url.database):
        raise AirflowConfigException(
            f"Cannot use relative path: `{conf.get('database', 'SQL_ALCHEMY_CONN')}` to connect to sqlite. "
            "Please use absolute path such as `sqlite:////tmp/airflow.db`."
        )

    if "SQLALCHEMY_ENGINE_OPTIONS" not in flask_app.config:
        flask_app.config["SQLALCHEMY_ENGINE_OPTIONS"] = settings.prepare_engine_args()

    csrf.init_app(flask_app)

    db = SQLAlchemy(flask_app)
    db.session = settings.Session

    configure_logging()
    configure_manifest_files(flask_app)
    init_api_auth(flask_app)

    with flask_app.app_context():
        init_appbuilder(flask_app, enable_plugins=enable_plugins)
        init_error_handlers(flask_app)
        # In two scenarios a Flask application can be created:
        # - To support Airflow 2 plugins relying on Flask (``enable_plugins`` is True)
        # - To support FAB auth manager (``enable_plugins`` is False)
        # There are some edge cases where ``enable_plugins`` is False but the auth manager configured is not
        # FAB auth manager. One example is ``run_update_fastapi_api_spec``, it calls
        # ``FabAuthManager().get_fastapi_app()`` to generate the openapi documentation regardless of the
        # configured auth manager.
        if enable_plugins:
            init_plugins(flask_app)
        elif isinstance(get_auth_manager(), FabAuthManager):
            init_api_auth_provider(flask_app)
            init_api_error_handlers(flask_app)
            init_airflow_session_interface(flask_app, db)
        init_jinja_globals(flask_app, enable_plugins=enable_plugins)
        init_wsgi_middleware(flask_app)
    return flask_app


def cached_app():
    """Return cached instance of Airflow WWW app."""
    global app
    if not app:
        app = create_app()
    return app


def purge_cached_app():
    """Remove the cached version of the app in global state."""
    global app
    app = None



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Create User and Role tables if missing.

Revision ID: 6709f7a774b9
Revises:
Create Date: 2024-09-03 17:06:38.040510

Note: This is a placeholder migration used to stamp the migration
when we create the migration from the ORM. Otherwise, it will run
without stamping the migration, leading to subsequent changes to
the tables not being migrated.
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6709f7a774b9"
down_revision = None
branch_labels = None
depends_on = None
fab_version = "1.4.0"


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "ab_group",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column("label", sa.String(length=150), nullable=True),
        sa.Column("description", sa.String(length=512), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_group_pkey")),
        sa.UniqueConstraint("name", name=op.f("ab_group_name_uq")),
        if_not_exists=True,
    )
    op.create_table(
        "ab_permission",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_permission_pkey")),
        sa.UniqueConstraint("name", name=op.f("ab_permission_name_uq")),
        if_not_exists=True,
    )
    op.create_table(
        "ab_register_user",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("first_name", sa.String(length=256), nullable=False),
        sa.Column("last_name", sa.String(length=256), nullable=False),
        sa.Column(
            "username",
            sa.String(length=512).with_variant(sa.String(length=512, collation="NOCASE"), "sqlite"),
            nullable=False,
        ),
        sa.Column("password", sa.String(length=256), nullable=True),
        sa.Column("email", sa.String(length=512), nullable=False),
        sa.Column("registration_date", sa.DateTime(), nullable=True),
        sa.Column("registration_hash", sa.String(length=256), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_register_user_pkey")),
        sa.UniqueConstraint("username", name=op.f("ab_register_user_username_uq")),
        if_not_exists=True,
    )
    op.create_table(
        "ab_role",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_role_pkey")),
        sa.UniqueConstraint("name", name=op.f("ab_role_name_uq")),
        if_not_exists=True,
    )
    op.create_table(
        "ab_user",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("first_name", sa.String(length=256), nullable=False),
        sa.Column("last_name", sa.String(length=256), nullable=False),
        sa.Column(
            "username",
            sa.String(length=512).with_variant(sa.String(length=512, collation="NOCASE"), "sqlite"),
            nullable=False,
        ),
        sa.Column("password", sa.String(length=256), nullable=True),
        sa.Column("active", sa.Boolean(), nullable=True),
        sa.Column("email", sa.String(length=512), nullable=False),
        sa.Column("last_login", sa.DateTime(), nullable=True),
        sa.Column("login_count", sa.Integer(), nullable=True),
        sa.Column("fail_login_count", sa.Integer(), nullable=True),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.Column("changed_on", sa.DateTime(), nullable=True),
        sa.Column("created_by_fk", sa.Integer(), nullable=True),
        sa.Column("changed_by_fk", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["changed_by_fk"], ["ab_user.id"], name=op.f("ab_user_changed_by_fk_fkey")),
        sa.ForeignKeyConstraint(["created_by_fk"], ["ab_user.id"], name=op.f("ab_user_created_by_fk_fkey")),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_user_pkey")),
        sa.UniqueConstraint("email", name=op.f("ab_user_email_uq")),
        sa.UniqueConstraint("username", name=op.f("ab_user_username_uq")),
        if_not_exists=True,
    )
    op.create_table(
        "ab_view_menu",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=250), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_view_menu_pkey")),
        sa.UniqueConstraint("name", name=op.f("ab_view_menu_name_uq")),
        if_not_exists=True,
    )
    op.create_table(
        "ab_group_role",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("group_id", sa.Integer(), nullable=True),
        sa.Column("role_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["group_id"], ["ab_group.id"], name=op.f("ab_group_role_group_id_fkey"), ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["role_id"], ["ab_role.id"], name=op.f("ab_group_role_role_id_fkey"), ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_group_role_pkey")),
        sa.UniqueConstraint("group_id", "role_id", name=op.f("ab_group_role_group_id_role_id_uq")),
        if_not_exists=True,
    )
    with op.batch_alter_table("ab_group_role", schema=None) as batch_op:
        batch_op.create_index("idx_group_id", ["group_id"], unique=False, if_not_exists=True)
        batch_op.create_index("idx_group_role_id", ["role_id"], unique=False, if_not_exists=True)

    op.create_table(
        "ab_permission_view",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("permission_id", sa.Integer(), nullable=True),
        sa.Column("view_menu_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["permission_id"], ["ab_permission.id"], name=op.f("ab_permission_view_permission_id_fkey")
        ),
        sa.ForeignKeyConstraint(
            ["view_menu_id"], ["ab_view_menu.id"], name=op.f("ab_permission_view_view_menu_id_fkey")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_permission_view_pkey")),
        sa.UniqueConstraint(
            "permission_id", "view_menu_id", name=op.f("ab_permission_view_permission_id_view_menu_id_uq")
        ),
        if_not_exists=True,
    )
    op.create_table(
        "ab_user_group",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=True),
        sa.Column("group_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["group_id"], ["ab_group.id"], name=op.f("ab_user_group_group_id_fkey"), ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["user_id"], ["ab_user.id"], name=op.f("ab_user_group_user_id_fkey"), ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_user_group_pkey")),
        sa.UniqueConstraint("user_id", "group_id", name=op.f("ab_user_group_user_id_group_id_uq")),
        if_not_exists=True,
    )
    with op.batch_alter_table("ab_user_group", schema=None) as batch_op:
        batch_op.create_index("idx_user_group_id", ["group_id"], unique=False, if_not_exists=True)
        batch_op.create_index("idx_user_id", ["user_id"], unique=False, if_not_exists=True)

    op.create_table(
        "ab_user_role",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=True),
        sa.Column("role_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["role_id"], ["ab_role.id"], name=op.f("ab_user_role_role_id_fkey"), ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["user_id"], ["ab_user.id"], name=op.f("ab_user_role_user_id_fkey"), ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_user_role_pkey")),
        sa.UniqueConstraint("user_id", "role_id", name=op.f("ab_user_role_user_id_role_id_uq")),
        if_not_exists=True,
    )
    op.create_table(
        "ab_permission_view_role",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("permission_view_id", sa.Integer(), nullable=True),
        sa.Column("role_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["permission_view_id"],
            ["ab_permission_view.id"],
            name=op.f("ab_permission_view_role_permission_view_id_fkey"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["role_id"], ["ab_role.id"], name=op.f("ab_permission_view_role_role_id_fkey"), ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("ab_permission_view_role_pkey")),
        sa.UniqueConstraint(
            "permission_view_id",
            "role_id",
            name=op.f("ab_permission_view_role_permission_view_id_role_id_uq"),
        ),
        if_not_exists=True,
    )
    # ### end Alembic commands ###


def downgrade() -> None: ...



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import os
from typing import TYPE_CHECKING

import pytest
import time_machine
from fastapi.testclient import TestClient

from airflow.api_fastapi.app import create_app
from airflow.api_fastapi.auth.managers.simple.user import SimpleAuthManagerUser

from tests_common.test_utils.config import conf_vars

if TYPE_CHECKING:
    from airflow.api_fastapi.auth.managers.simple.simple_auth_manager import SimpleAuthManager


API_PATHS = {
    "public": "/api/v2",
    "ui": "/ui",
}


BASE_URL = "http://testserver"


def get_api_path(request):
    """Determine the API path based on the test's subdirectory."""
    test_dir = os.path.dirname(request.path)
    subdirectory_name = test_dir.split("/")[-1]

    return API_PATHS.get(subdirectory_name, "/")


@pytest.fixture
def test_client(request):
    with conf_vars(
        {
            (
                "core",
                "auth_manager",
            ): "airflow.api_fastapi.auth.managers.simple.simple_auth_manager.SimpleAuthManager",
        }
    ):
        app = create_app()
        auth_manager: SimpleAuthManager = app.state.auth_manager
        # set time_very_before to 2014-01-01 00:00:00 and time_very_after to tomorrow
        # to make the JWT token always valid for all test cases with time_machine
        time_very_before = datetime.datetime(2014, 1, 1, 0, 0, 0)
        time_after = datetime.datetime.now() + datetime.timedelta(days=1)
        with time_machine.travel(time_very_before, tick=False):
            token = auth_manager._get_token_signer(
                expiration_time_in_seconds=(time_after - time_very_before).total_seconds()
            ).generate(
                auth_manager.serialize_user(SimpleAuthManagerUser(username="test", role="admin")),
            )
        yield TestClient(
            app, headers={"Authorization": f"Bearer {token}"}, base_url=f"{BASE_URL}{get_api_path(request)}"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""User sub-commands."""

from __future__ import annotations

import functools
import getpass
import json
import os
import random
import re
import string
from typing import Any

from marshmallow import Schema, fields, validate
from marshmallow.exceptions import ValidationError

from airflow.cli.simple_table import AirflowConsole
from airflow.cli.utils import print_export_output
from airflow.providers.fab.auth_manager.cli_commands.utils import get_application_builder
from airflow.utils import cli as cli_utils
from airflow.utils.cli import suppress_logs_and_warning
from airflow.utils.providers_configuration_loader import providers_configuration_loaded


class UserSchema(Schema):
    """user collection item schema."""

    id = fields.Int()
    firstname = fields.Str(required=True)
    lastname = fields.Str(required=True)
    username = fields.Str(required=True)
    email = fields.Email(required=True)
    roles = fields.List(fields.Str, required=True, validate=validate.Length(min=1))


@suppress_logs_and_warning
@providers_configuration_loaded
def users_list(args):
    """List users at the command line."""
    with get_application_builder() as appbuilder:
        users = appbuilder.sm.get_all_users()
        fields = ["id", "username", "email", "first_name", "last_name", "roles"]

        AirflowConsole().print_as(
            data=users, output=args.output, mapper=lambda x: {f: x.__getattribute__(f) for f in fields}
        )


@cli_utils.action_cli(check_db=True)
@providers_configuration_loaded
def users_create(args):
    """Create new user in the DB."""
    with get_application_builder() as appbuilder:
        role = appbuilder.sm.find_role(args.role)
        if not role:
            valid_roles = appbuilder.sm.get_all_roles()
            raise SystemExit(f"{args.role} is not a valid role. Valid roles are: {valid_roles}")
        password = _create_password(args)
        if appbuilder.sm.find_user(args.username):
            print(f"{args.username} already exist in the db")
            return
        user = appbuilder.sm.add_user(
            args.username, args.firstname, args.lastname, args.email, role, password
        )
        if user:
            print(f'User "{args.username}" created with role "{args.role}"')
        else:
            raise SystemExit("Failed to create user")


def _find_user(args):
    if not args.username and not args.email:
        raise SystemExit("Missing args: must supply one of --username or --email")

    if args.username and args.email:
        raise SystemExit("Conflicting args: must supply either --username or --email, but not both")

    with get_application_builder() as appbuilder:
        user = appbuilder.sm.find_user(username=args.username, email=args.email)
        if not user:
            raise SystemExit(f'User "{args.username or args.email}" does not exist')
    return user


@cli_utils.action_cli
@providers_configuration_loaded
def user_reset_password(args):
    """Reset user password user from DB."""
    user = _find_user(args)
    password = _create_password(args)
    with get_application_builder() as appbuilder:
        if appbuilder.sm.reset_password(user.id, password):
            print(f'User "{user.username}" password reset successfully')
        else:
            raise SystemExit("Failed to reset user password")


def _create_password(args):
    if args.use_random_password:
        password = "".join(random.choices(string.printable, k=16))
    elif args.password:
        password = args.password
    else:
        password = getpass.getpass("Password:")
        password_confirmation = getpass.getpass("Repeat for confirmation:")
        if password != password_confirmation:
            raise SystemExit("Passwords did not match")
    return password


@cli_utils.action_cli
@providers_configuration_loaded
def users_delete(args):
    """Delete user from DB."""
    user = _find_user(args)

    # Clear the associated user roles first.
    user.roles.clear()

    with get_application_builder() as appbuilder:
        if appbuilder.sm.del_register_user(user):
            print(f'User "{user.username}" deleted')
        else:
            raise SystemExit("Failed to delete user")


@cli_utils.action_cli
@providers_configuration_loaded
def users_manage_role(args, remove=False):
    """Delete or appends user roles."""
    with get_application_builder() as appbuilder:
        user = _find_user(args)
        role = appbuilder.sm.find_role(args.role)
        if not role:
            valid_roles = appbuilder.sm.get_all_roles()
            raise SystemExit(f'"{args.role}" is not a valid role. Valid roles are: {valid_roles}')

        if remove:
            if role not in user.roles:
                raise SystemExit(f'User "{user.username}" is not a member of role "{args.role}"')

            user.roles = [r for r in user.roles if r != role]
            appbuilder.sm.update_user(user)
            print(f'User "{user.username}" removed from role "{args.role}"')
        else:
            if role in user.roles:
                raise SystemExit(f'User "{user.username}" is already a member of role "{args.role}"')

            user.roles.append(role)
            appbuilder.sm.update_user(user)
            print(f'User "{user.username}" added to role "{args.role}"')


@providers_configuration_loaded
def users_export(args):
    """Export all users to the json file."""
    with get_application_builder() as appbuilder:
        users = appbuilder.sm.get_all_users()
        fields = ["id", "username", "email", "first_name", "last_name", "roles"]

        # In the User model the first and last name fields have underscores,
        # but the corresponding parameters in the CLI don't
        def remove_underscores(s):
            return re.sub("_", "", s)

        users = [
            {
                remove_underscores(field): user.__getattribute__(field)
                if field != "roles"
                else [r.name for r in user.roles]
                for field in fields
            }
            for user in users
        ]

        with open(args.export, "w") as file:
            file.write(json.dumps(users, sort_keys=True, indent=4))
            print_export_output("users", users, file)


@cli_utils.action_cli
@providers_configuration_loaded
def users_import(args):
    """Import users from the json file."""
    json_file = getattr(args, "import")
    if not os.path.exists(json_file):
        raise SystemExit(f"File '{json_file}' does not exist")

    users_list = None
    try:
        with open(json_file) as file:
            users_list = json.loads(file.read())
    except ValueError as e:
        raise SystemExit(f"File '{json_file}' is not valid JSON. Error: {e}")

    users_created, users_updated = _import_users(users_list)
    if users_created:
        users_created_str = "\n\t".join(users_created)
        print(f"Created the following users:\n\t{users_created_str}")

    if users_updated:
        users_updated_str = "\n\t".join(users_updated)
        print(f"Updated the following users:\n\t{users_updated_str}")


def _import_users(users_list: list[dict[str, Any]]):
    with get_application_builder() as appbuilder:
        users_created = []
        users_updated = []

        try:
            UserSchema(many=True).load(users_list)
        except ValidationError as e:
            msg = []
            for row_num, failure in e.normalized_messages().items():
                msg.append(f"[Item {row_num}]")
                for key, value in failure.items():
                    msg.append(f"\t{key}: {value}")
            msg_str = "\n".join(msg)
            raise SystemExit(f"Error: Input file didn't pass validation. See below:\n{msg_str}")

        for user in users_list:
            roles = []
            for rolename in user["roles"]:
                role = appbuilder.sm.find_role(rolename)
                if not role:
                    valid_roles = appbuilder.sm.get_all_roles()
                    raise SystemExit(
                        f'Error: "{rolename}" is not a valid role. Valid roles are: {valid_roles}'
                    )
                roles.append(role)

            existing_user = appbuilder.sm.find_user(email=user["email"])
            if existing_user:
                print(f"Found existing user with email '{user['email']}'")
                if existing_user.username != user["username"]:
                    raise SystemExit(
                        f"Error: Changing the username is not allowed - "
                        f"please delete and recreate the user with email {user['email']!r}"
                    )

                existing_user.roles = roles
                existing_user.first_name = user["firstname"]
                existing_user.last_name = user["lastname"]
                appbuilder.sm.update_user(existing_user)
                users_updated.append(user["email"])
            else:
                print(f"Creating new user with email '{user['email']}'")
                appbuilder.sm.add_user(
                    username=user["username"],
                    first_name=user["firstname"],
                    last_name=user["lastname"],
                    email=user["email"],
                    role=roles,
                )

                users_created.append(user["email"])

        return users_created, users_updated


add_role = functools.partial(users_manage_role, remove=False)
remove_role = functools.partial(users_manage_role, remove=True)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from airflow.api_fastapi.app import get_auth_manager
from airflow.configuration import conf
from airflow.providers.fab.www.security.permissions import (
    ACTION_CAN_ACCESS_MENU,
    ACTION_CAN_CREATE,
    ACTION_CAN_DELETE,
    ACTION_CAN_EDIT,
    ACTION_CAN_READ,
)

if TYPE_CHECKING:
    try:
        from airflow.api_fastapi.auth.managers.base_auth_manager import ExtendedResourceMethod
    except ImportError:
        from airflow.api_fastapi.auth.managers.base_auth_manager import (
            ResourceMethod as ExtendedResourceMethod,
        )
    from airflow.providers.fab.auth_manager.fab_auth_manager import FabAuthManager

# Convert methods to FAB action name
_MAP_METHOD_NAME_TO_FAB_ACTION_NAME: dict[ExtendedResourceMethod, str] = {
    "POST": ACTION_CAN_CREATE,
    "GET": ACTION_CAN_READ,
    "PUT": ACTION_CAN_EDIT,
    "DELETE": ACTION_CAN_DELETE,
    "MENU": ACTION_CAN_ACCESS_MENU,
}
log = logging.getLogger(__name__)


def get_session_lifetime_config():
    """Get session timeout configs and handle outdated configs gracefully."""
    session_lifetime_minutes = conf.get("fab", "session_lifetime_minutes", fallback=None)
    minutes_per_day = 24 * 60
    if not session_lifetime_minutes:
        session_lifetime_days = 30
        session_lifetime_minutes = minutes_per_day * session_lifetime_days

    log.debug("User session lifetime is set to %s minutes.", session_lifetime_minutes)

    return int(session_lifetime_minutes)


def get_fab_auth_manager() -> FabAuthManager:
    from airflow.providers.fab.auth_manager.fab_auth_manager import FabAuthManager

    auth_manager = get_auth_manager()
    if not isinstance(auth_manager, FabAuthManager):
        raise RuntimeError(
            "This functionality is only available with if FabAuthManager is configured as auth manager in the environment."
        )
    return auth_manager


def get_fab_action_from_method_map():
    """Return the map associating a method to a FAB action."""
    return _MAP_METHOD_NAME_TO_FAB_ACTION_NAME


def get_method_from_fab_action_map():
    """Return the map associating a FAB action to a method."""
    return {
        **{v: k for k, v in _MAP_METHOD_NAME_TO_FAB_ACTION_NAME.items()},
    }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.fab.auth_manager.models import User

pytestmark = pytest.mark.db_test


@pytest.mark.parametrize(
    "user_id, expected_id",
    [
        (999, "999"),
        ("999", "999"),
    ],
)
def test_get_id_returns_str(user_id: int | str, expected_id: str) -> None:
    """
    Ensure get_id() always returns a string representation of the id.
    """
    user = User()
    user.id = user_id
    result = user.get_id()
    assert isinstance(result, str), f"Expected str, got {type(result)}"
    assert result == expected_id



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from sqlalchemy import Column, MetaData, String, Table

from airflow.providers.fab.auth_manager.models import (
    add_index_on_ab_register_user_username_postgres,
    add_index_on_ab_user_username_postgres,
)

_mock_conn = mock.MagicMock()
_mock_conn.dialect = mock.MagicMock()
_mock_conn.dialect.name = "postgresql"


def test_add_index_on_ab_user_username_postgres():
    table = Table("test_table", MetaData(), Column("username", String))

    assert len(table.indexes) == 0

    add_index_on_ab_user_username_postgres(table, _mock_conn)

    # Assert that the index was added to the table
    assert len(table.indexes) == 1

    add_index_on_ab_user_username_postgres(table, _mock_conn)

    # Assert that index is not re-added when the schema is recreated
    assert len(table.indexes) == 1


def test_add_index_on_ab_register_user_username_postgres():
    table = Table("test_table", MetaData(), Column("username", String))

    assert len(table.indexes) == 0

    add_index_on_ab_register_user_username_postgres(table, _mock_conn)

    # Assert that the index was added to the table
    assert len(table.indexes) == 1

    add_index_on_ab_register_user_username_postgres(table, _mock_conn)

    # Assert that index is not re-added when the schema is recreated
    assert len(table.indexes) == 1



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest
from sqlalchemy import select

from airflow.providers.fab.auth_manager.models import Role
from airflow.providers.fab.auth_manager.security_manager.override import EXISTING_ROLES
from airflow.providers.fab.www.api_connexion.exceptions import EXCEPTIONS_LINK_MAP
from airflow.providers.fab.www.security import permissions

from unit.fab.auth_manager.api_endpoints.api_connexion_utils import (
    assert_401,
    create_role,
    create_user,
    delete_role,
    delete_user,
)

pytestmark = pytest.mark.db_test


@pytest.fixture
def configured_app(minimal_app_for_auth_api):
    app = minimal_app_for_auth_api
    with app.app_context():
        create_user(
            app,
            username="test",
            role_name="Test",
            permissions=[
                (permissions.ACTION_CAN_CREATE, permissions.RESOURCE_ROLE),
                (permissions.ACTION_CAN_READ, permissions.RESOURCE_ROLE),
                (permissions.ACTION_CAN_EDIT, permissions.RESOURCE_ROLE),
                (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_ROLE),
                (permissions.ACTION_CAN_READ, permissions.RESOURCE_ACTION),
            ],
        )
        create_user(app, username="test_no_permissions", role_name="TestNoPermissions")
        yield app

        delete_user(app, username="test")
        delete_user(app, username="test_no_permissions")
        session = app.appbuilder.session
        existing_roles = set(EXISTING_ROLES)
        existing_roles.update(["Test", "TestNoPermissions"])
        roles = session.scalars(select(Role).where(~Role.name.in_(existing_roles))).unique().all()
        for role in roles:
            delete_role(app, role.name)


class TestRoleEndpoint:
    @pytest.fixture(autouse=True)
    def setup_attrs(self, configured_app) -> None:
        self.app = configured_app
        self.client = self.app.test_client()


class TestGetRoleEndpoint(TestRoleEndpoint):
    def test_should_response_200(self):
        response = self.client.get("/fab/v1/roles/Admin", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        assert response.json["name"] == "Admin"

    def test_should_respond_404(self):
        response = self.client.get("/fab/v1/roles/invalid-role", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 404
        assert response.json == {
            "detail": "Role with name 'invalid-role' was not found",
            "status": 404,
            "title": "Role not found",
            "type": EXCEPTIONS_LINK_MAP[404],
        }

    def test_should_raises_401_unauthenticated(self):
        response = self.client.get("/fab/v1/roles/Admin")
        assert_401(response)

    def test_should_raise_403_forbidden(self):
        response = self.client.get(
            "/fab/v1/roles/Admin", environ_overrides={"REMOTE_USER": "test_no_permissions"}
        )
        assert response.status_code == 403

    @pytest.mark.parametrize(
        "set_auth_role_public, expected_status_code",
        (("Public", 403), ("Admin", 200)),
        indirect=["set_auth_role_public"],
    )
    def test_with_auth_role_public_set(self, set_auth_role_public, expected_status_code):
        response = self.client.get("/fab/v1/roles/Admin")
        assert response.status_code == expected_status_code, response.json


class TestGetRolesEndpoint(TestRoleEndpoint):
    def test_should_response_200(self):
        response = self.client.get("/fab/v1/roles", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        existing_roles = set(EXISTING_ROLES)
        existing_roles.update(["Test", "TestNoPermissions"])
        assert response.json["total_entries"] == len(existing_roles)
        roles = {role["name"] for role in response.json["roles"]}
        assert roles == existing_roles

    def test_should_raises_401_unauthenticated(self):
        response = self.client.get("/fab/v1/roles")
        assert_401(response)

    def test_should_raises_400_for_invalid_order_by(self):
        response = self.client.get(
            "/fab/v1/roles?order_by=invalid", environ_overrides={"REMOTE_USER": "test"}
        )
        assert response.status_code == 400
        msg = "Ordering with 'invalid' is disallowed or the attribute does not exist on the model"
        assert response.json["detail"] == msg

    def test_should_raise_403_forbidden(self):
        response = self.client.get("/fab/v1/roles", environ_overrides={"REMOTE_USER": "test_no_permissions"})
        assert response.status_code == 403

    @pytest.mark.parametrize(
        "set_auth_role_public, expected_status_code",
        (("Public", 403), ("Admin", 200)),
        indirect=["set_auth_role_public"],
    )
    def test_with_auth_role_public_set(self, set_auth_role_public, expected_status_code):
        response = self.client.get("/fab/v1/roles")
        assert response.status_code == expected_status_code, response.json


class TestGetRolesEndpointPaginationandFilter(TestRoleEndpoint):
    @pytest.mark.parametrize(
        "url, expected_roles",
        [
            ("/fab/v1/roles?limit=1", ["Admin"]),
            ("/fab/v1/roles?limit=2", ["Admin", "Op"]),
            (
                "/fab/v1/roles?offset=1",
                ["Op", "Public", "Test", "TestNoPermissions", "User", "Viewer"],
            ),
            (
                "/fab/v1/roles?offset=0",
                ["Admin", "Op", "Public", "Test", "TestNoPermissions", "User", "Viewer"],
            ),
            ("/fab/v1/roles?limit=1&offset=2", ["Public"]),
            ("/fab/v1/roles?limit=1&offset=1", ["Op"]),
            (
                "/fab/v1/roles?limit=2&offset=2",
                ["Public", "Test"],
            ),
        ],
    )
    def test_can_handle_limit_and_offset(self, url, expected_roles):
        response = self.client.get(url, environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        existing_roles = set(EXISTING_ROLES)
        existing_roles.update(["Test", "TestNoPermissions"])
        assert response.json["total_entries"] == len(existing_roles)
        roles = [role["name"] for role in response.json["roles"] if role]

        assert roles == expected_roles


class TestGetPermissionsEndpoint(TestRoleEndpoint):
    def test_should_response_200(self):
        response = self.client.get("/fab/v1/permissions", environ_overrides={"REMOTE_USER": "test"})
        actions = {i[0] for i in self.app.appbuilder.sm.get_all_permissions() if i}
        assert response.status_code == 200
        assert response.json["total_entries"] == len(actions)
        returned_actions = {perm["name"] for perm in response.json["actions"]}
        assert actions == returned_actions

    def test_should_raises_401_unauthenticated(self):
        response = self.client.get("/fab/v1/permissions")
        assert_401(response)

    def test_should_raise_403_forbidden(self):
        response = self.client.get(
            "/fab/v1/permissions", environ_overrides={"REMOTE_USER": "test_no_permissions"}
        )
        assert response.status_code == 403

    @pytest.mark.parametrize(
        "set_auth_role_public, expected_status_code",
        (("Public", 403), ("Admin", 200)),
        indirect=["set_auth_role_public"],
    )
    def test_with_auth_role_public_set(self, set_auth_role_public, expected_status_code):
        response = self.client.get("/fab/v1/permissions")
        assert response.status_code == expected_status_code, response.json


class TestPostRole(TestRoleEndpoint):
    def test_post_should_respond_200(self):
        payload = {
            "name": "Test2",
            "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
        }
        response = self.client.post("/fab/v1/roles", json=payload, environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        role = self.app.appbuilder.sm.find_role("Test2")
        assert role is not None

    @pytest.mark.parametrize(
        "payload, error_message",
        [
            (
                {
                    "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
                },
                "{'name': ['Missing data for required field.']}",
            ),
            (
                {
                    "name": "TestRole",
                    "actionss": [
                        {
                            "resource": {"name": "Connections"},  # actionss not correct
                            "action": {"name": "can_create"},
                        }
                    ],
                },
                "{'actionss': ['Unknown field.']}",
            ),
            (
                {
                    "name": "TestRole",
                    "actions": [
                        {
                            "resources": {"name": "Connections"},  # resources is invalid, should be resource
                            "action": {"name": "can_create"},
                        }
                    ],
                },
                "{'actions': {0: {'resources': ['Unknown field.']}}}",
            ),
            (
                {
                    "name": "TestRole",
                    "actions": [
                        {"resource": {"name": "Connections"}, "actions": {"name": "can_create"}}
                    ],  # actions is invalid, should be action
                },
                "{'actions': {0: {'actions': ['Unknown field.']}}}",
            ),
            (
                {
                    "name": "TestRole",
                    "actions": [
                        {
                            "resource": {"name": "FooBars"},  # FooBars is not a resource
                            "action": {"name": "can_create"},
                        }
                    ],
                },
                "The specified resource: 'FooBars' was not found",
            ),
            (
                {
                    "name": "TestRole",
                    "actions": [
                        {"resource": {"name": "Connections"}, "action": {"name": "can_amend"}}
                    ],  # can_amend is not an action
                },
                "The specified action: 'can_amend' was not found",
            ),
        ],
    )
    def test_post_should_respond_400_for_invalid_payload(self, payload, error_message):
        response = self.client.post("/fab/v1/roles", json=payload, environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 400
        assert response.json == {
            "detail": error_message,
            "status": 400,
            "title": "Bad Request",
            "type": EXCEPTIONS_LINK_MAP[400],
        }

    def test_post_should_respond_409_already_exist(self):
        payload = {
            "name": "Test",
            "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
        }
        response = self.client.post("/fab/v1/roles", json=payload, environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 409
        assert response.json == {
            "detail": "Role with name 'Test' already exists; please update with the PATCH endpoint",
            "status": 409,
            "title": "Conflict",
            "type": EXCEPTIONS_LINK_MAP[409],
        }

    def test_should_raises_401_unauthenticated(self):
        response = self.client.post(
            "/fab/v1/roles",
            json={
                "name": "Test2",
                "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
            },
        )

        assert_401(response)

    def test_should_raise_403_forbidden(self):
        response = self.client.post(
            "/fab/v1/roles",
            json={
                "name": "mytest2",
                "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
            },
            environ_overrides={"REMOTE_USER": "test_no_permissions"},
        )
        assert response.status_code == 403

    @pytest.mark.parametrize(
        "set_auth_role_public, expected_status_code",
        (("Public", 403), ("Admin", 200)),
        indirect=["set_auth_role_public"],
    )
    def test_with_auth_role_public_set(self, set_auth_role_public, expected_status_code):
        payload = {
            "name": "Test2",
            "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
        }
        response = self.client.post("/fab/v1/roles", json=payload)
        assert response.status_code == expected_status_code, response.json


class TestDeleteRole(TestRoleEndpoint):
    def test_delete_should_respond_204(self, session):
        role = create_role(self.app, "mytestrole")
        response = self.client.delete(f"/fab/v1/roles/{role.name}", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 204
        role_obj = session.scalars(select(Role).where(Role.name == role.name)).all()
        assert len(role_obj) == 0

    def test_delete_should_respond_404(self):
        response = self.client.delete(
            "/fab/v1/roles/invalidrolename", environ_overrides={"REMOTE_USER": "test"}
        )
        assert response.status_code == 404
        assert response.json == {
            "detail": "Role with name 'invalidrolename' was not found",
            "status": 404,
            "title": "Role not found",
            "type": EXCEPTIONS_LINK_MAP[404],
        }

    def test_should_raises_401_unauthenticated(self):
        response = self.client.delete("/fab/v1/roles/test")

        assert_401(response)

    def test_should_raise_403_forbidden(self):
        response = self.client.delete(
            "/fab/v1/roles/test", environ_overrides={"REMOTE_USER": "test_no_permissions"}
        )
        assert response.status_code == 403

    @pytest.mark.parametrize(
        "set_auth_role_public, expected_status_code",
        (("Public", 403), ("Admin", 204)),
        indirect=["set_auth_role_public"],
    )
    def test_with_auth_role_public_set(self, set_auth_role_public, expected_status_code):
        role = create_role(self.app, "mytestrole")
        response = self.client.delete(f"/fab/v1/roles/{role.name}")
        assert response.status_code == expected_status_code, response.location


class TestPatchRole(TestRoleEndpoint):
    @pytest.mark.parametrize(
        "payload, expected_name, expected_actions",
        [
            ({"name": "mytest"}, "mytest", []),
            (
                {
                    "name": "mytest2",
                    "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
                },
                "mytest2",
                [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
            ),
        ],
    )
    def test_patch_should_respond_200(self, payload, expected_name, expected_actions):
        role = create_role(self.app, "mytestrole")
        response = self.client.patch(
            f"/fab/v1/roles/{role.name}", json=payload, environ_overrides={"REMOTE_USER": "test"}
        )
        assert response.status_code == 200
        assert response.json["name"] == expected_name
        assert response.json["actions"] == expected_actions

    def test_patch_should_update_correct_roles_permissions(self):
        create_role(self.app, "role_to_change")
        create_role(self.app, "already_exists")

        response = self.client.patch(
            "/fab/v1/roles/role_to_change",
            json={
                "name": "already_exists",
                "actions": [{"action": {"name": "can_delete"}, "resource": {"name": "XComs"}}],
            },
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200

        updated_permissions = self.app.appbuilder.sm.find_role("role_to_change").permissions
        assert len(updated_permissions) == 1
        assert updated_permissions[0].resource.name == "XComs"
        assert updated_permissions[0].action.name == "can_delete"

        assert len(self.app.appbuilder.sm.find_role("already_exists").permissions) == 0

    @pytest.mark.parametrize(
        "update_mask, payload, expected_name, expected_actions",
        [
            (
                "?update_mask=name",
                {
                    "name": "mytest2",
                    "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
                },
                "mytest2",
                [],
            ),
            (
                "?update_mask=name, actions",  # both name and actions in update mask
                {
                    "name": "mytest2",
                    "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
                },
                "mytest2",
                [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
            ),
        ],
    )
    def test_patch_should_respond_200_with_update_mask(
        self, update_mask, payload, expected_name, expected_actions
    ):
        role = create_role(self.app, "mytestrole")
        assert role.permissions == []
        response = self.client.patch(
            f"/fab/v1/roles/{role.name}{update_mask}",
            json=payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200
        assert response.json["name"] == expected_name
        assert response.json["actions"] == expected_actions

    def test_patch_should_respond_400_for_invalid_fields_in_update_mask(self):
        role = create_role(self.app, "mytestrole")
        payload = {"name": "testme"}
        response = self.client.patch(
            f"/fab/v1/roles/{role.name}?update_mask=invalid_name",
            json=payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 400
        assert response.json["detail"] == "'invalid_name' in update_mask is unknown"

    @pytest.mark.parametrize(
        "payload, expected_error",
        [
            (
                {
                    "name": "testme",
                    "permissions": [  # Using permissions instead of actions should raise
                        {"resource": {"name": "Connections"}, "action": {"name": "can_create"}}
                    ],
                },
                "{'permissions': ['Unknown field.']}",
            ),
            (
                {
                    "name": "testme",
                    "actions": [
                        {
                            "view_menu": {"name": "Connections"},  # Using view_menu instead of resource
                            "action": {"name": "can_create"},
                        }
                    ],
                },
                "{'actions': {0: {'view_menu': ['Unknown field.']}}}",
            ),
            (
                {
                    "name": "testme",
                    "actions": [
                        {
                            "resource": {"name": "FooBars"},  # Using wrong resource name
                            "action": {"name": "can_create"},
                        }
                    ],
                },
                "The specified resource: 'FooBars' was not found",
            ),
            (
                {
                    "name": "testme",
                    "actions": [
                        {
                            "resource": {"name": "Connections"},  # Using wrong action name
                            "action": {"name": "can_invalid"},
                        }
                    ],
                },
                "The specified action: 'can_invalid' was not found",
            ),
        ],
    )
    def test_patch_should_respond_400_for_invalid_update(self, payload, expected_error):
        role = create_role(self.app, "mytestrole")
        response = self.client.patch(
            f"/fab/v1/roles/{role.name}",
            json=payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 400
        assert response.json["detail"] == expected_error

    def test_should_raises_401_unauthenticated(self):
        response = self.client.patch(
            "/fab/v1/roles/test",
            json={
                "name": "mytest2",
                "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
            },
        )

        assert_401(response)

    def test_should_raise_403_forbidden(self):
        response = self.client.patch(
            "/fab/v1/roles/test",
            json={
                "name": "mytest2",
                "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
            },
            environ_overrides={"REMOTE_USER": "test_no_permissions"},
        )
        assert response.status_code == 403

    @pytest.mark.parametrize(
        "set_auth_role_public, expected_status_code",
        (("Public", 403), ("Admin", 200)),
        indirect=["set_auth_role_public"],
    )
    def test_with_auth_role_public_set(self, set_auth_role_public, expected_status_code):
        role = create_role(self.app, "mytestrole")
        response = self.client.patch(
            f"/fab/v1/roles/{role.name}",
            json={"name": "mytest"},
        )
        assert response.status_code == expected_status_code, response.json



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from importlib import reload
from unittest import mock

import pytest

from airflow.cli import cli_parser

from tests_common.test_utils.config import conf_vars

pytestmark = [pytest.mark.db_test]
try:
    from airflow.providers.fab.auth_manager.cli_commands import db_command
    from airflow.providers.fab.auth_manager.models.db import FABDBManager

    class TestFABCLiDB:
        @classmethod
        def setup_class(cls):
            with conf_vars(
                {
                    (
                        "core",
                        "auth_manager",
                    ): "airflow.providers.fab.auth_manager.fab_auth_manager.FabAuthManager",
                }
            ):
                # Reload the module to use FAB auth manager
                reload(cli_parser)
                # Clearing the cache before calling it
                cli_parser.get_parser.cache_clear()
                cls.parser = cli_parser.get_parser()

        @mock.patch.object(FABDBManager, "resetdb")
        def test_cli_resetdb(self, mock_resetdb):
            db_command.resetdb(self.parser.parse_args(["fab-db", "reset", "--yes"]))

            mock_resetdb.assert_called_once_with(skip_init=False)

        @mock.patch.object(FABDBManager, "resetdb")
        def test_cli_resetdb_skip_init(self, mock_resetdb):
            db_command.resetdb(self.parser.parse_args(["fab-db", "reset", "--yes", "--skip-init"]))
            mock_resetdb.assert_called_once_with(skip_init=True)

        @pytest.mark.parametrize(
            "args, called_with",
            [
                (
                    [],
                    dict(
                        to_revision=None,
                        from_revision=None,
                        show_sql_only=False,
                    ),
                ),
                (
                    ["--show-sql-only"],
                    dict(
                        to_revision=None,
                        from_revision=None,
                        show_sql_only=True,
                    ),
                ),
                (
                    ["--to-revision", "abc"],
                    dict(
                        to_revision="abc",
                        from_revision=None,
                        show_sql_only=False,
                    ),
                ),
                (
                    ["--to-revision", "abc", "--show-sql-only"],
                    dict(to_revision="abc", from_revision=None, show_sql_only=True),
                ),
                (
                    ["--to-revision", "abc", "--from-revision", "abc123", "--show-sql-only"],
                    dict(
                        to_revision="abc",
                        from_revision="abc123",
                        show_sql_only=True,
                    ),
                ),
            ],
        )
        @mock.patch.object(FABDBManager, "upgradedb")
        def test_cli_upgrade_success(self, mock_upgradedb, args, called_with):
            db_command.migratedb(self.parser.parse_args(["fab-db", "migrate", *args]))
            mock_upgradedb.assert_called_once_with(**called_with)

        @pytest.mark.parametrize(
            "args, pattern",
            [
                pytest.param(
                    ["--to-revision", "abc", "--to-version", "1.3.0"],
                    "Cannot supply both",
                    id="to both version and revision",
                ),
                pytest.param(
                    ["--from-revision", "abc", "--from-version", "1.3.0"],
                    "Cannot supply both",
                    id="from both version and revision",
                ),
                pytest.param(["--to-version", "1.2.0"], "Unknown version '1.2.0'", id="unknown to version"),
                pytest.param(["--to-version", "abc"], "Invalid version 'abc'", id="invalid to version"),
                pytest.param(
                    ["--to-revision", "abc", "--from-revision", "abc123"],
                    "used with `--show-sql-only`",
                    id="requires offline",
                ),
                pytest.param(
                    ["--to-revision", "abc", "--from-version", "1.3.0"],
                    "used with `--show-sql-only`",
                    id="requires offline",
                ),
                pytest.param(
                    ["--to-revision", "abc", "--from-version", "1.1.25", "--show-sql-only"],
                    "Unknown version '1.1.25'",
                    id="unknown from version",
                ),
                pytest.param(
                    ["--to-revision", "adaf", "--from-version", "abc", "--show-sql-only"],
                    "Invalid version 'abc'",
                    id="invalid from version",
                ),
            ],
        )
        @mock.patch.object(FABDBManager, "upgradedb")
        def test_cli_migratedb_failure(self, mock_upgradedb, args, pattern):
            with pytest.raises(SystemExit, match=pattern):
                db_command.migratedb(self.parser.parse_args(["fab-db", "migrate", *args]))
except (ModuleNotFoundError, ImportError):
    pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from sqlalchemy import Table

from airflow.exceptions import AirflowException
from airflow.utils.db import initdb
from airflow.utils.db_manager import RunDBManager

from tests_common.test_utils.config import conf_vars

pytestmark = [pytest.mark.db_test]


class TestRunDBManagerWithFab:
    @conf_vars(
        {("database", "external_db_managers"): "airflow.providers.fab.auth_manager.models.db.FABDBManager"}
    )
    def test_db_manager_uses_config(self):
        from airflow.providers.fab.auth_manager.models.db import FABDBManager

        run_db_manager = RunDBManager()
        assert run_db_manager._managers == [FABDBManager]

    @conf_vars(
        {("database", "external_db_managers"): "airflow.providers.fab.auth_manager.models.db.FABDBManager"}
    )
    def test_defining_table_same_name_as_airflow_table_name_raises(self):
        from sqlalchemy import Column, Integer, String

        run_db_manager = RunDBManager()
        metadata = run_db_manager._managers[0].metadata
        # Add dag_run table to metadata
        mytable = Table(
            "dag_run", metadata, Column("id", Integer, primary_key=True), Column("name", String(50))
        )
        metadata._add_table("dag_run", None, mytable)
        with pytest.raises(AirflowException, match="Table 'dag_run' already exists in the Airflow metadata"):
            run_db_manager.validate()
        metadata._remove_table("dag_run", None)

    @mock.patch.object(RunDBManager, "upgradedb")
    @mock.patch.object(RunDBManager, "initdb")
    def test_init_db_calls_rundbmanager(self, mock_initdb, mock_upgrade_db, session):
        initdb(session=session)
        mock_initdb.assert_called()
        mock_initdb.assert_called_once_with(session)

    @conf_vars(
        {("database", "external_db_managers"): "airflow.providers.fab.auth_manager.models.db.FABDBManager"}
    )
    @mock.patch("airflow.providers.fab.auth_manager.models.db.FABDBManager")
    def test_rundbmanager_calls_dbmanager_methods(self, mock_fabdb_manager, session):
        mock_fabdb_manager.supports_table_dropping = True
        fabdb_manager = mock_fabdb_manager.return_value
        ext_db = RunDBManager()
        # initdb
        ext_db.initdb(session=session)
        fabdb_manager.initdb.assert_called_once()
        # upgradedb
        ext_db.upgradedb(session=session)
        fabdb_manager.upgradedb.assert_called_once()
        # drop_tables
        connection = mock.MagicMock()
        ext_db.drop_tables(session, connection)
        mock_fabdb_manager.return_value.drop_tables.assert_called_once_with(connection)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Session authentication backend."""

from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar, cast

from flask import Response

from airflow.api_fastapi.app import get_auth_manager

CLIENT_AUTH: tuple[str, str] | Any | None = None


def init_app(_):
    """Initialize authentication backend."""


T = TypeVar("T", bound=Callable)


def requires_authentication(function: T):
    """Decorate functions that require authentication."""

    @wraps(function)
    def decorated(*args, **kwargs):
        if not get_auth_manager().is_logged_in():
            return Response("Unauthorized", 401, {})
        return function(*args, **kwargs)

    return cast("T", decorated)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from urllib.parse import unquote, urljoin, urlsplit

from flask import (
    g,
    make_response,
    redirect,
    render_template,
    request,
    url_for,
)
from flask_appbuilder import IndexView, expose

from airflow.api_fastapi.app import get_auth_manager
from airflow.api_fastapi.auth.managers.base_auth_manager import COOKIE_NAME_JWT_TOKEN
from airflow.configuration import conf
from airflow.providers.fab.version_compat import AIRFLOW_V_3_1_1_PLUS

# Following the release of https://github.com/python/cpython/issues/102153 in Python 3.9.17 on
# June 6, 2023, we are adding extra sanitization of the urls passed to get_safe_url method to make it works
# the same way regardless if the user uses latest Python patchlevel versions or not. This also follows
# a recommended solution by the Python core team.
#
# From: https://github.com/python/cpython/commit/d28bafa2d3e424b6fdcfd7ae7cde8e71d7177369
#
#   We recommend that users of these APIs where the values may be used anywhere
#   with security implications code defensively. Do some verification within your
#   code before trusting a returned component part.  Does that ``scheme`` make
#   sense?  Is that a sensible ``path``?  Is there anything strange about that
#   ``hostname``?  etc.
#
# C0 control and space to be stripped per WHATWG spec.
# == "".join([chr(i) for i in range(0, 0x20 + 1)])
_WHATWG_C0_CONTROL_OR_SPACE = (
    "\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c"
    "\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f "
)


class FabIndexView(IndexView):
    """
    A simple view that inherits from FAB index view.

    The only goal of this view is to redirect the user to the Airflow 3 UI index page if the user is
    authenticated. It is impossible to redirect the user directly to the Airflow 3 UI index page before
    redirecting them to this page because FAB itself defines the logic redirection and does not allow external
    redirect.
    """

    @expose("/")
    def index(self):
        if g.user is not None and g.user.is_authenticated:
            token = get_auth_manager().generate_jwt(g.user)
            response = make_response(redirect(f"{conf.get('api', 'base_url', fallback='/')}", code=302))

            secure = request.scheme == "https" or bool(conf.get("api", "ssl_cert", fallback=""))
            # In Airflow 3.1.1 authentication changes, front-end no longer handle the token
            # See https://github.com/apache/airflow/pull/55506
            if AIRFLOW_V_3_1_1_PLUS:
                response.set_cookie(COOKIE_NAME_JWT_TOKEN, token, secure=secure, httponly=True)
            else:
                response.set_cookie(COOKIE_NAME_JWT_TOKEN, token, secure=secure)

            return response
        return redirect(conf.get("api", "base_url", fallback="/"), code=302)


def show_traceback(error):
    """Show Traceback for a given error."""
    return render_template("airflow/traceback.html"), 500


def not_found(error):
    """Show Not Found on screen for any error in the Webserver."""
    return (
        render_template(
            "airflow/error.html",
            hostname="",
            status_code=404,
            error_message="Page cannot be found.",
        ),
        404,
    )


def get_safe_url(url):
    """Given a user-supplied URL, ensure it points to our web server."""
    if not url:
        return url_for("FabIndexView.index")

    # If the url contains semicolon, redirect it to homepage to avoid
    # potential XSS. (Similar to https://github.com/python/cpython/pull/24297/files (bpo-42967))
    if ";" in unquote(url):
        return url_for("FabIndexView.index")

    url = url.lstrip(_WHATWG_C0_CONTROL_OR_SPACE)

    host_url = urlsplit(request.host_url)
    redirect_url = urlsplit(urljoin(request.host_url, url))
    if not (redirect_url.scheme in ("http", "https") and host_url.netloc == redirect_url.netloc):
        return url_for("FabIndexView.index")

    # This will ensure we only redirect to the right scheme/netloc
    return redirect_url.geturl()


def method_not_allowed(error):
    """Show Method Not Allowed on screen for any error in the Webserver."""
    return (
        render_template(
            "airflow/error.html",
            status_code=405,
            error_message="Received an invalid request.",
        ),
        405,
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from pathlib import Path

from diagrams import Cluster, Diagram, Edge
from diagrams.custom import Custom
from diagrams.onprem.client import User
from diagrams.onprem.database import PostgreSQL
from rich.console import Console

MY_DIR = Path(__file__).parent
MY_FILENAME = Path(__file__).with_suffix("").name
PYTHON_MULTIPROCESS_LOGO = MY_DIR.parents[1] / "diagrams" / "python_multiprocess_logo.png"

console = Console(width=400, color_system="standard")


def generate_fab_auth_manager_airflow_diagram():
    image_file = (MY_DIR / MY_FILENAME).with_suffix(".png")
    console.print(f"[bright_blue]Generating architecture image {image_file}")
    with Diagram(
        name="",
        show=False,
        direction="LR",
        curvestyle="ortho",
        filename=MY_FILENAME,
    ):
        user = User("User")
        with Cluster("Airflow environment"):
            webserver = Custom("Webserver(s)", PYTHON_MULTIPROCESS_LOGO.as_posix())

            with Cluster("FAB provider"):
                fab_auth_manager = Custom("FAB auth manager", PYTHON_MULTIPROCESS_LOGO.as_posix())
            with Cluster("Core Airflow"):
                auth_manager_interface = Custom(
                    "Auth manager\ninterface", PYTHON_MULTIPROCESS_LOGO.as_posix()
                )

            db = PostgreSQL("Metadata DB")

        user >> Edge(color="black", style="solid", reverse=True, label="Access to the console") >> webserver
        (
            webserver
            >> Edge(color="black", style="solid", reverse=True, label="Is user authorized?")
            >> fab_auth_manager
        )
        (fab_auth_manager >> Edge(color="black", style="solid", reverse=True) >> db)
        (
            fab_auth_manager
            >> Edge(color="black", style="dotted", reverse=False, label="Inherit")
            >> auth_manager_interface
        )

    console.print(f"[green]Generating architecture image {image_file}")


if __name__ == "__main__":
    generate_fab_auth_manager_airflow_diagram()
